import os
import jinja2

# the files are in `data/evaluation_utils`
# evaluation_utils is a symlink to data/evaluation_utils for convenience
import pandas as pd

from evaluation_utils.mnist_evaluator import MNISTEvaluator

loader = jinja2.FileSystemLoader("jinja_templates")
env = jinja2.Environment(autoescape=True, loader=loader)


class Constants:
    SHARED_DISK = os.getenv("AICROWD_SHARED_DIR", "test_debug_data/shared")
    DATASET_DIR = os.getenv("AICROWD_DATASET_DIR", "test_debug_data/data")
    GROUND_TRUTH_DIR = os.getenv(
        "AICROWD_GROUND_TRUTH_DIR", "test_debug_data/ground_truth"
    )

    PREDICTIONS_FILE_PATH = os.path.join(SHARED_DISK, "predictions.csv")
    GROUND_TRUTH_PATH = os.path.join(GROUND_TRUTH_DIR, "labels.csv")


class AIcrowdEvaluator:
    def __init__(self, **kwargs):
        self.evaluator = MNISTEvaluator(
            test_data_path=Constants.DATASET_DIR,
            ground_truth_data_path=Constants.GROUND_TRUTH_PATH,
            predictions_file_path=Constants.PREDICTIONS_FILE_PATH,
        )
    
    def live_evaluate(self, ground_truth_path: str, predictions_file_path: str) -> dict:
        # Not compulsory but encouraged to do so.
        # For long running evaluations, this will help the participants understand the performance
        # of their models before the completion of the evaluation.
        return self.evaluator.scoring(ground_truth_path, predictions_file_path, partial_score=True)

    def render_current_status_as_markdown(self) -> str:
        """This method updates the GitLab issue page comment with the current evaluation
        progress. The content returned should be in markdown.

        This method is invoked from a different process isolated from the
        participant's code.
        """
        def state_emoji(progress: float) -> str:
            if progress == 0:
                return ":hourglass:"
            elif progress < 100:
                return ":stopwatch:"
            return ":white_check_mark:"

        env.filters["state_emoji"] = state_emoji
        # Load the jinja template from jinja_templates/status-v1.md
        j2_template = env.get_template("status-v1.md")

        try:
            # Get evaluation progress based on predictions completed
            total_image_count = len(pd.read_csv(Constants.GROUND_TRUTH_PATH))
            predictions = pd.read_csv(Constants.PREDICTIONS_FILE_PATH)
            progress_percent = len(predictions) * 100 / total_image_count

            # Get live scores for predictions so far
            scores = self.live_evaluate(
                Constants.GROUND_TRUTH_PATH,
                Constants.PREDICTIONS_FILE_PATH,
            )
        except:
            progress_percent = 0
            scores = {
                "score": 0,
                "score_secondary": 0,
            }

        return j2_template.render(
            progress_percent=progress_percent,
            f1_score=round(scores["score"], 4),
            accuracy_score=round(scores["score_secondary"], 4),
        )

    def evaluate(self):
        """For information on how the scores should be organized, adding private
        scores, etc., please refer
        https://wiki.aicrowd.com/share/efa47829-ef4e-452e-b75e-c5ac3fb24290
        """
        scores = self.evaluator.scoring(
            ground_truth_path=Constants.GROUND_TRUTH_PATH,
            prediction_file_path=Constants.PREDICTIONS_FILE_PATH,
        )
        return scores


if __name__ == "__main__":
    evaluator = AIcrowdEvaluator()
    score = evaluator.evaluate()
    print(score)

    print(evaluator.render_current_status_as_markdown())
